from .default_settings import complete_settings_dict
from .validate import get_default_value_from_schema
from copy import deepcopy
from math import log2
from .charts import load_chart_definition, altair_if_installed_else_json


class ComparisonColumn:
    def __init__(self, column_dict):
        self.column_dict = column_dict

    def __getitem__(self, i):
        return self.column_dict[i]

    def _dict_key_else_default_value(self, key):
        cd = self.column_dict

        if key in cd:
            return cd[key]
        else:
            return get_default_value_from_schema(key, True)

    @property
    def custom_comparison(self):
        cd = self.column_dict
        if "custom_name" in cd:
            return True
        elif "col_name" in cd:
            return False

    @property
    def columns_used(self):
        cd = self.column_dict
        if "col_name" in cd:
            return [cd["col_name"]]
        elif "custom_name" in cd:
            return cd["custom_columns_used"]

    @property
    def name(self):
        cd = self.column_dict
        if "custom_name" in cd:
            return cd["custom_name"]
        elif "col_name" in cd:
            return cd["col_name"]

    @property
    def gamma_name(self):
        return f"gamma_{self.name}"

    @property
    def num_levels(self):
        cd = self.column_dict
        m_probs = cd["m_probabilities"]
        u_probs = cd["u_probabilities"]
        if len(m_probs) == len(u_probs):
            return len(m_probs)
        else:
            raise ValueError("Length of m and u probs unequal")

    @property
    def max_gamma_index(self):
        return self.num_levels - 1

    @property
    def input_cols_used(self):
        cd = self.column_dict
        if "custom_name" in cd:
            return cd["custom_columns_used"]
        elif "col_name" in cd:
            return [cd["col_name"]]

    @property
    def term_frequency_adjustments(self):
        cd = self.column_dict
        return  cd["term_frequency_adjustments"]


    def get_m_u_bayes_at_gamma_index(self, gamma_index):

        # if -1 this indicates a null field
        if gamma_index == -1:
            m = 1
            u = 1
        else:
            m = self["m_probabilities"][gamma_index]
            u = self["u_probabilities"][gamma_index]
        if u != 0 and m is not None and u is not None:
            bayes = m / u
            if m != 0:
                log_2_bayes = log2(bayes)
            else:
                log_2_bayes = None
        else:
            bayes = None
            log_2_bayes = None
        return {
            "m_probability": m,
            "u_probability": u,
            "bayes_factor": bayes,
            "log2_bayes_factor": log_2_bayes,
        }

    def describe_row_dict(self, row_dict, proportion_of_matches=None):

        gamma_index = int(row_dict[self.gamma_name])
        row_desc = self.level_as_dict(gamma_index, proportion_of_matches)

        row_desc["value_l"] = ", ".join(
            [str(row_dict[c + "_l"]) for c in self.columns_used]
        )
        row_desc["value_r"] = ", ".join(
            [str(row_dict[c + "_r"]) for c in self.columns_used]
        )

        return row_desc

    def set_m_probability(self, level: int, prob: float, force: bool = False):
        cd = self.column_dict
        fixed = self._dict_key_else_default_value("fix_m_probabilities")
        if not fixed or force:
            cd["m_probabilities"][level] = prob

    def set_u_probability(self, level: int, prob: float, force: bool = False):
        cd = self.column_dict
        fixed = self._dict_key_else_default_value("fix_u_probabilities")
        if not fixed or force:
            cd["u_probabilities"][level] = prob

    def reset_probabilities(self, force: bool = False):
        cd = self.column_dict
        fixed_m = self._dict_key_else_default_value("fix_m_probabilities")
        fixed_u = self._dict_key_else_default_value("fix_u_probabilities")
        if not fixed_m or force:
            if "m_probabilities" in cd:
                cd["m_probabilities"] = [None for c in cd["m_probabilities"]]

        if not fixed_u or force:
            if "u_probabilities" in cd:
                cd["u_probabilities"] = [None for c in cd["u_probabilities"]]

    def level_as_dict(self, gamma_index, proportion_of_matches=None):

        d = self.get_m_u_bayes_at_gamma_index(gamma_index)

        d["gamma_column_name"] = f"gamma_{self.name}"
        d["level_name"] = f"level_{gamma_index}"

        d["gamma_index"] = gamma_index
        d["column_name"] = self.name
        d["max_gamma_index"] = self.max_gamma_index
        d["num_levels"] = self.num_levels

        d["level_proportion"] = None
        if proportion_of_matches:
            lam = proportion_of_matches
            m = d["m_probability"]
            u = d["u_probability"]
            # Check they both not None
            if m and u:
                d["level_proportion"] = m * lam + u * (1 - lam)
            else:
                d["level_proportion"] = None

        return d

    def as_rows(self, proportion_of_matches=None):
        """Convert to rows e.g. to use to plot
        in a chart"""
        rows = []
        for gamma_index in range(self.num_levels):
            r = self.level_as_dict(gamma_index, proportion_of_matches)
            rows.append(r)
        return rows

    def _repr_pretty_(self, p, cycle):

        p.text("------------------------------------")
        p.break_()
        p.text(f"Comparison of {self.name}")
        p.break_()
        for row in self.as_rows():
            p.text(f"{row['level_name']}")
            p.break_()
            value_m = (
                f"{row['m_probability']*100:.3g}%" if row["m_probability"] else "None"
            )
            p.text(f"   Proportion in level amongst matches:       {value_m}")
            p.break_()
            value_u = (
                f"{row['u_probability']*100:.3g}%" if row["u_probability"] else "None"
            )
            p.text(f"   Proportion in level amongst non-matches:   {value_u}%")
            p.break_()
            value_b = (
                f"{row['bayes_factor']*100:.3g}%" if row["bayes_factor"] else "None"
            )
            p.text(f"   Bayes factor:                              {value_b}")
            p.break_()


class Settings:
    def __init__(self, settings_dict):
        self.settings_dict = deepcopy(settings_dict)

    def __getitem__(self, i):
        return self.settings_dict[i]

    def __setitem__(self, key, value):
        self.settings_dict[key] = value

    def complete_settings_dict(self, spark):
        """Complete all fields in the setting dictionary
        taking values from defaults"""
        self.settings_dict = complete_settings_dict(self.settings_dict, spark)

    @property
    def comparison_column_dict(self):
        sd = self.settings_dict
        lookup = {}
        for i, c in enumerate(sd["comparison_columns"]):
            c["gamma_index"] = i
            cc = ComparisonColumn(c)
            name = cc.name
            lookup[name] = cc
        return lookup

    @property
    def comparison_columns_list(self):
        return list(self.comparison_column_dict.values())

    @property
    def any_cols_have_tf_adjustments(self):
        return any([c.term_frequency_adjustments for c in self.comparison_columns_list])

    def get_comparison_column(self, col_name_or_custom_name):
        if col_name_or_custom_name in self.comparison_column_dict:
            return self.comparison_column_dict[col_name_or_custom_name]
        else:
            raise KeyError(
                f"You requested comparison column {col_name_or_custom_name}"
                " but it does not exist in the settings object"
            )

    def reset_all_probabilities(self, force: bool = False):
        sd = self.settings_dict
        sd["proportion_of_matches"] = None
        for c in self.comparison_columns_list:
            c.reset_probabilities(force=force)

    def m_u_as_rows(self):
        """Convert to rows e.g. to use to plot
        in a chart"""
        rows = []
        for c in self.comparison_columns_list:
            rows.extend(c.as_rows(self["proportion_of_matches"]))
        return rows

    def set_m_probability(
        self, name: str, level: int, prob: float, force: bool = False
    ):
        cc = self.get_comparison_column(name)
        cc.set_m_probability(level, prob, force)

    def set_u_probability(
        self, name: str, level: int, prob: float, force: bool = False
    ):
        cc = self.get_comparison_column(name)
        cc.set_u_probability(level, prob, force)

    def overwrite_m_u_probs_from_other_settings_dict(
        self, incoming_settings_dict, overwrite_m=True, overwrite_u=True
    ):
        """Overwrite the m and u probabilities with
        the values from another settings dict where comparison column exists
        in this settings object
        """
        incoming_settings_obj = Settings(incoming_settings_dict)

        for cc_incoming in incoming_settings_obj.comparison_columns_list:
            if cc_incoming.name in self.comparison_column_dict:
                cc_existing = self.get_comparison_column(cc_incoming.name)

                if overwrite_m:
                    if "m_probabilities" in cc_incoming.column_dict:
                        cc_existing.column_dict["m_probabilities"] = cc_incoming[
                            "m_probabilities"
                        ]

                if overwrite_u:
                    if "u_probabilities" in cc_incoming.column_dict:
                        cc_existing.column_dict["u_probabilities"] = cc_incoming[
                            "u_probabilities"
                        ]

    def remove_comparison_column(self, name):
        removed = False
        new_ccs = []
        for cc in self.comparison_columns_list:
            if cc.name != name:
                new_ccs.append(cc.column_dict)
            else:
                removed = True
        self.settings_dict["comparison_columns"] = new_ccs
        if not removed:
            raise ValueError(f"Could not find a column named {name}")

    def probability_distribution_chart(self):  # pragma: no cover
        chart_path = "probability_distribution_chart.json"
        chart = load_chart_definition(chart_path)
        chart["data"]["values"] = self.m_u_as_rows()
        chart["title"]["subtitle"] += f" {self['proportion_of_matches']:.3g}"

        return altair_if_installed_else_json(chart)

    def bayes_factor_chart(self):  # pragma: no cover
        chart_path = "bayes_factor_chart_def.json"
        chart = load_chart_definition(chart_path)
        chart["data"]["values"] = self.m_u_as_rows()
        return altair_if_installed_else_json(chart)

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        if cycle:
            p.text("ComparisonColumn()")
        else:
            value = (
                f"{self['proportion_of_matches']:.4g}"
                if self["proportion_of_matches"]
                else "None"
            )
            p.text(f"λ (proportion of matches) = {value}")
            for c in self.comparison_columns_list:
                p.break_()
                p.pretty(c)



from textwrap import dedent

sqlparse_exists = True
try:
    import sqlparse
except ImportError:
    sqlparse_exists = False


def _format_sql(sql):
    if sqlparse_exists:
        return sqlparse.format(sql, reindent=True, keyword_case="upper")
    else:
        return dedent(sql)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splink']

package_data = \
{'': ['*'],
 'splink': ['files/chart_defs/*',
            'files/external_js/*',
            'files/settings_jsonschema.json',
            'files/templates/*',
            'jars/scala-udf-similarity-0.0.9.jar']}

install_requires = \
['jsonschema>=3.2,<4.0', 'typeguard>=2.10.0,<3.0.0']

setup_kwargs = {
    'name': 'splink',
    'version': '2.0.1',
    'description': "Implementation in Apache Spark of the EM algorithm to estimate parameters of Fellegi-Sunter's canonical model of record linkage.",
    'long_description': "![image](https://user-images.githubusercontent.com/7570107/85285114-3969ac00-b488-11ea-88ff-5fca1b34af1f.png)\n\n[![Coverage Status](https://coveralls.io/repos/github/moj-analytical-services/splink/badge.svg?branch=master)](https://coveralls.io/github/moj-analytical-services/splink?branch=master)\n![issues-status](https://img.shields.io/github/issues-raw/moj-analytical-services/splink)\n![python-version-dependency](https://img.shields.io/badge/python-%3E%3D3.6-blue)\n\n# splink: Probabilistic record linkage and deduplication at scale\n\n`splink` implements Fellegi-Sunter's canonical model of record linkage in Apache Spark, including the EM algorithm to estimate parameters of the model.\n\nIt:\n\n- Works at much greater scale than current open source implementations (100 million records+).\n\n- Runs quickly - with runtimes of less than an hour.\n\n- Has a highly transparent methodology; match scores can be easily explained both graphically and in words\n\n- Is highly accurate\n\nIt is assumed that users of Splink are familiar with the probabilistic record linkage theory, and the Fellegi-Sunter model in particular. A [series of interactive articles](https://www.robinlinacre.com/probabilistic_linkage/) explores the theory behind Splink.\n\nThe statistical model behind `splink` is the same as that used in the R [fastLink package](https://github.com/kosukeimai/fastLink). Accompanying the fastLink package is an [academic paper](http://imai.fas.harvard.edu/research/files/linkage.pdf) that describes this model. This is the best place to start for users wanting to understand the theory about how `splink` works.\n\n[Data Matching](https://link.springer.com/book/10.1007/978-3-642-31164-2), a book by Peter Christen, is another excellent resource.\n\n## Installation\n\n`splink` is a Python package. It uses the Spark Python API to execute data linking jobs in a Spark cluster. It has been tested in Apache Spark 2.3 and 2.4.\n\nInstall splink using:\n\n`pip install splink`\n\nNote that Splink requires `pyspark` and a working Spark installation. These are not specified as explicit dependencies becuase it is assumed users have an existing pyspark setup they wish to use.\n\n## Interactive demo\n\nYou can run demos of `splink` in an interactive Jupyter notebook by clicking the button below:\n\n[![Binder](https://mybinder.org/badge.svg)](https://mybinder.org/v2/gh/moj-analytical-services/splink_demos/master?urlpath=lab/tree/index.ipynb)\n\n## Documentation\n\nThe best documentation is currently a series of demonstrations notebooks in the [splink_demos](https://github.com/moj-analytical-services/splink_demos) repo.\n\n## Other tools in the Splink family\n\n### Splink Graph\n\n[`splink_graph`](https://github.com/moj-analytical-services/splink_graph) is a graph utility library for use in Apache Spark. It computes graph metrics on the outputs of data linking. The repo is [here](<(https://github.com/moj-analytical-services/splink_graph)>)\n\n- Quality assurance of linkage results and identifying false positive links\n- Computing quality metrics associated with groups (clusters) of linked records\n- Automatically identifying possible false positive links in clusters\n\n### Splink Cluster Studio\n\n[`splink_cluster_studio`](http://github.com/moj-analytical-services/splink_cluster_studio) creates an interactive html dashboard from Splink output that allows you to visualise and analyse a sample of clusters from your record linkage. The repo is [here](http://github.com/moj-analytical-services/splink_cluster_studio).\n\n### Splink Synthetic Data\n\nThis [code](https://github.com/moj-analytical-services/splink_synthetic_data) is able to generate realistic test datasets for linkage using the WikiData Query Service.\n\nIt has been used to [performance test the accuracy of various Splink models](https://www.robinlinacre.com/comparing_splink_models/).\n\n### Interactive settings editor with autocomplete\n\nWe also provide an interactive `splink` settings editor and example settings [here](https://moj-analytical-services.github.io/splink_settings_editor/).\n\n### Starting parameter generation tools\n\nA tool to generate custom `m` and `u` probabilities can be found [here](https://www.robinlinacre.com/m_u_generator/).\n\n## Blog\n\nYou can read a short blog post about `splink` [here](https://robinlinacre.com/introducing_splink/).\n\n## Videos\n\nYou can find a short video introducing `splink` and running though an introductory demo [here](https://www.youtube.com/watch?v=_8lV2Lbd6Xs&feature=youtu.be&t=1295).\n\nA 'best practices and performance tuning' tutorial can be found [here](https://www.youtube.com/watch?v=HzcqrRvXhCE).\n\n## Acknowledgements\n\nWe are very grateful to [ADR UK](https://www.adruk.org/) (Administrative Data Research UK) for providing funding for this work as part of the [Data First](https://www.adruk.org/our-work/browse-all-projects/data-first-harnessing-the-potential-of-linked-administrative-data-for-the-justice-system-169/) project.\n\nWe are also very grateful to colleagues at the UK's Office for National Statistics for their expert advice and peer review of this work.\n",
    'author': 'Robin Linacre',
    'author_email': 'robinlinacre@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/moj-analytical-services/splink',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from functools import reduce
from copy import copy
import warnings

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from typeguard import typechecked

from .charts import load_chart_definition, altair_if_installed_else_json


def _equal_spaced_buckets(num_buckets, extent):
    buckets = [x for x in range(num_buckets + 1)]
    span = extent[1] - extent[0]
    buckets = [extent[0] + span * (x / num_buckets) for x in buckets]
    return buckets


@typechecked
def _calc_probability_density(
    df_e: DataFrame,
    spark: SparkSession,
    buckets=None,
    score_colname="match_probability",
    symmetric=True,
):

    """perform splink score histogram calculations / internal function

    Compute a histogram using the provided buckets.

        Args:
            df_e (DataFrame): A dataframe of record comparisons containing a
                splink score, e.g. as produced by the expectation step
            spark (SparkSession): SparkSession object
            score_colname: is the score in another column? defaults to match_probability.  also try match_weight
            buckets: accepts either a list of split points or an integer number that is used
                to create equally spaced split points.  It defaults to 100 equally
                spaced split points
            symmetric : if True then the histogram is symmetric

        Returns:
            (list) : list of rows of histogram bins for appropriate splink score variable ready to be plotted.
    """

    if score_colname == "match_probability":
        extent = (0.0, 1.0)
    else:
        weight_max = df_e.agg({score_colname: "max"}).collect()[0][0]
        weight_min = df_e.agg({score_colname: "min"}).collect()[0][0]
        extent = (weight_min, weight_max)
        if symmetric:
            extent_max = max(abs(weight_max), abs(weight_min))
            extent = (-extent_max, extent_max)

    # if buckets a list then use it. if None... then create default. if integer then create equal bins
    if isinstance(buckets, int) and buckets != 0:
        buckets = _equal_spaced_buckets(buckets, extent)
    elif buckets is None:
        buckets = _equal_spaced_buckets(100, extent)

    buckets.sort()

    # ensure bucket splits are in ascending order
    if score_colname == "match_probability":
        if buckets[0] != 0:
            buckets = [0.0] + buckets

        if buckets[-1] != 1.0:
            buckets = buckets + [1.0]

    hist = df_e.select(score_colname).rdd.flatMap(lambda x: x).histogram(buckets)

    # get bucket from and to points
    bin_low = hist[0]
    bin_high = copy(hist[0])
    bin_low.pop()
    bin_high.pop(0)
    counts = hist[1]

    rows = []
    for item in zip(bin_low, bin_high, counts):
        new_row = {
            "splink_score_bin_low": item[0],
            "splink_score_bin_high": item[1],
            "count_rows": item[2],
        }
        rows.append(new_row)

    for r in rows:
        r["binwidth"] = r["splink_score_bin_high"] - r["splink_score_bin_low"]
        r["freqdensity"] = r["count_rows"] / r["binwidth"]

    sumfreqdens = reduce(lambda a, b: a + b["freqdensity"], rows, 0)

    for r in rows:
        r["normalised"] = r["freqdensity"] / sumfreqdens

    return rows


def _create_probability_density_plot(data):
    """plot score histogram

    Args:
        data (list): A list of rows of histogram bins
            as produced by the _calc_probability_density function
    Returns:
        if altair is installed a plot. if altair is not installed
            then it returns the vega lite chart spec as a dictionary
    """

    hist_def_dict = load_chart_definition("score_histogram.json")
    hist_def_dict["data"]["values"] = data

    return altair_if_installed_else_json(hist_def_dict)


def splink_score_histogram(
    df_e: DataFrame,
    spark: SparkSession,
    buckets=None,
    score_colname=None,
    symmetric=True,
):

    """splink score histogram diagnostic plot public API function

    Compute a histogram using the provided buckets and plot the result.

    Args:
        df_e (DataFrame): A dataframe of record comparisons containing a splink score,
            e.g. as produced by the `get_scored_comparisons` function
        spark (SparkSession): SparkSession object
        score_colname : is the score in another column? defaults to None
        buckets : accepts either a list of split points or an integer number that is used to
            create equally spaced split points. It defaults to 100 equally spaced split points from 0.0 to 1.0
        symmetric : if True then the histogram is symmetric
     Returns:
        if altair library is installed this function returns a histogram plot. if altair is not installed
        then it returns the vega lite chart spec as a dictionary
    """

    rows = _calc_probability_density(
        df_e,
        spark=spark,
        buckets=buckets,
        score_colname=score_colname,
        symmetric=symmetric,
    )

    return _create_probability_density_plot(rows)
